from dagster import DefaultScheduleStatus, ScheduleDefinition

from dagster_meltano.utils import generate_dagster_name


class Schedule:
    def __init__(self, meltano_schedule: dict) -> None:
        self.name = meltano_schedule["name"]
        self.cron_interval = meltano_schedule["cron_interval"]
        self.job_name = meltano_schedule["job"]["name"]

    @property
    def dagster_name(self) -> str:
        return generate_dagster_name(self.name)

    @property
    def dagster_job_name(self) -> str:
        return generate_dagster_name(self.job_name)

    @property
    def dagster_schedule(self) -> ScheduleDefinition:
        return ScheduleDefinition(
            name=self.dagster_name,
            job_name=self.dagster_job_name,
            cron_schedule=self.cron_interval,
            default_status=DefaultScheduleStatus.RUNNING,
        )



"""Meltano Dagster extension."""
from __future__ import annotations

import os
import pkgutil
import subprocess
import sys
import time
from pathlib import Path
from typing import Any

import files_dagster_ext
import structlog
import typer
from cookiecutter.main import cookiecutter
from meltano.edk import models
from meltano.edk.extension import ExtensionBase
from meltano.edk.process import Invoker, log_subprocess_error
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm, Prompt

log = structlog.get_logger()

spinner = Progress(
    SpinnerColumn(),
    TextColumn("[progress.description]{task.description}"),
    transient=True,
)


class Dagster(ExtensionBase):
    """Extension implementing the ExtensionBase interface."""

    def __init__(self) -> None:
        """Initialize the extension."""
        self.invokers = {
            "dagster": Invoker("dagster"),
            "dagit": Invoker("dagit"),
            "meltano": Invoker("meltano", env={**os.environ, "MELTANO_ENVIRONMENT": ""}),
        }

    @property
    def utility_name(self) -> str:
        return os.getenv("MELTANO_UTILITY_NAME", "dagster")

    @property
    def cookiecutter_template_dir(self) -> str:
        """
        Find the location of the files extention.
        This houses the cookiecutter project.
        """
        return Path(files_dagster_ext.__path__._path[0]) / "dagster"

    def initialize(self, force: bool = False) -> None:
        # TODO: This method needs to be cleaned up in the future.
        repository_dir = Prompt.ask(
            "Where do you want to install the Dagster project?",
            default="$MELTANO_PROJECT_ROOT/orchestrate",
        )

        with spinner as progress:
            progress.add_task(description="Setting Dagster repository directory", total=None)
            self.get_invoker_by_name("meltano").run(
                "config",
                self.utility_name,
                "set",
                "repository_dir",
                repository_dir,
            )

        repository_dir = os.path.expandvars(repository_dir)
        repository_dir = Path(repository_dir)

        install_github_actions = Confirm.ask(
            "Do you want to install Github Actions to deploy to Dagster Cloud Serverless? (Does not work yet)",
            default=True,
        )

        dbt_installed = Confirm.ask(
            "Do you have DBT installed?",
            default=True,
        )

        if dbt_installed:
            dbt_plugin = Prompt.ask(
                "Which DBT plugin do you have installed?",
                choices=["postgres", "snowflake", "bigquery", "redshift"],
            )
            print(
                "[blue]Make sure you define the DBT environment variables in the Dagster utility.[/blue]"
            )
        else:
            dbt_plugin = None

        cookiecutter_config = {
            "project_name": repository_dir.name,
            "install_github_actions": install_github_actions,
            "dbt_plugin": dbt_plugin,
        }

        cookiecutter(
            template=str(self.cookiecutter_template_dir),
            no_input=True,
            extra_context=cookiecutter_config,
            output_dir=str(repository_dir.parent),
            overwrite_if_exists=True,
        )

        print("[green]Successfully initialized your Dagster project![/green]")
        print("[green]Start Dagit by running `meltano invoke dagster:up`[/green]")

    def invoke(self, invoker: Invoker, command_name: str | None, *command_args: Any) -> None:
        """Invoke the underlying cli, that is being wrapped by this extension.

        Args:
            command_name: The name of the command to invoke.
            command_args: The arguments to pass to the command.
        """
        try:
            invoker.run_and_log(command_name, *command_args)
        except subprocess.CalledProcessError as err:
            log_subprocess_error(f"{invoker.bin} {command_name}", err, "Dagster invocation failed")
            sys.exit(err.returncode)

    def describe(self) -> models.Describe:
        """Describe the extension.

        Returns:
            The extension description
        """
        # TODO: could we auto-generate all or portions of this from typer instead?
        return models.Describe(
            commands=[
                models.ExtensionCommand(name="dagster_extension", description="extension commands"),
                models.InvokerCommand(name="dagster_invoker", description="pass through invoker"),
            ]
        )

    def get_invoker_by_name(self, invoker_name: str) -> Invoker:
        return self.invokers[invoker_name]

    def pass_through_invoker(
        self, invoker_name: str, logger: structlog.BoundLogger, *command_args: Any
    ) -> None:
        """Pass-through invoker.

        This method is used to invoke the wrapped CLI with arbitrary arguments.
        Note this method will hard exit the process if an unhandled exception is
        encountered.

        Args:
            invoker_name: The name of the invoker, currently "dagster" or "dagit".
            logger: The logger to use in the event an exception needs to be logged.
            *command_args: The arguments to pass to the command.
        """
        logger.debug(
            "pass through invoker called",
            command_args=command_args,
        )
        try:
            self.pre_invoke(None)
        except Exception:
            logger.exception(
                "pre_invoke failed with uncaught exception, please report to maintainer"
            )
            sys.exit(1)

        try:
            invoker = self.get_invoker_by_name(invoker_name)
            self.invoke(invoker, None, command_args)
        except Exception:
            logger.exception("invoke failed with uncaught exception, please report to maintainer")
            sys.exit(1)

        try:
            self.post_invoke(None)
        except Exception:
            logger.exception(
                "post_invoke failed with uncaught exception, please report to maintainer"  # noqa: E501
            )
            sys.exit(1)


if __name__ == "__main__":
    path = files_dagster_ext.__path__



"""Passthrough shim for Dagster extension."""
import sys

import structlog
from meltano.edk.logging import pass_through_logging_config

from dagster_ext.extension import Dagster


def pass_through_cli_dagster() -> None:
    """Pass through CLI entry point for Dagster."""
    pass_through_logging_config()
    ext = Dagster()
    ext.pass_through_invoker(
        "dagster",
        structlog.getLogger("dagster_invoker"),
        *sys.argv[1:] if len(sys.argv) > 1 else [],
    )


def pass_through_cli_dagit() -> None:
    """Pass through CLI entry point for Dagit."""
    pass_through_logging_config()
    ext = Dagster()
    ext.pass_through_invoker(
        "dagit",
        structlog.getLogger("dagit_invoker"),
        *sys.argv[1:] if len(sys.argv) > 1 else [],
    )



from __future__ import annotations

import asyncio
import json
from functools import lru_cache
from typing import TYPE_CHECKING

from dagster import (
    In,
    JobDefinition,
    Nothing,
    OpDefinition,
    OpExecutionContext,
    get_dagster_logger,
    job,
    op,
)

from dagster_meltano.meltano_invoker import MeltanoInvoker
from dagster_meltano.utils import generate_dagster_name

if TYPE_CHECKING:
    from dagster_meltano.meltano_resource import MeltanoResource

dagster_logger = get_dagster_logger()


async def log_processor(reader: asyncio.streams.StreamReader, log_type: str) -> None:
    """Log the output of a stream.

    Args:
        reader: The stream reader to read from.
    """
    # TODO: Clean up the logging
    while True:
        if reader.at_eof():
            break
        data = await reader.readline()
        log_line_raw = data.decode("utf-8").rstrip()

        if not log_line_raw:
            continue

        try:
            log_line = json.loads(log_line_raw)

            if log_line.get("level") == "debug":
                dagster_logger.debug(log_line.get("event", log_line))
            else:
                dagster_logger.info(log_line.get("event", log_line))
        except json.decoder.JSONDecodeError:
            dagster_logger.info(log_line_raw)

        await asyncio.sleep(0)


@lru_cache
def task_op_factory(task: str) -> OpDefinition:
    """
    This factory is cached to make sure the same tasks can be reused in the
    same repository.
    """

    @op(
        name=generate_dagster_name(task),
        description=f"Run `{task}` using Meltano.",
        ins={"after": In(Nothing)},
        tags={"kind": "meltano"},
        required_resource_keys={"meltano"},
    )
    def dagster_op(context: OpExecutionContext):
        meltano_resource: MeltanoResource = context.resources.meltano
        meltano_resource.meltano_invoker.run_and_log(
            "run",
            log_processor,
            task.split(),
        )

    return dagster_op


class Job:
    def __init__(self, meltano_job: dict, meltano_invoker: MeltanoInvoker) -> None:
        self.name = meltano_job["job_name"]
        self.tasks = meltano_job["tasks"]
        self.meltano_invoker = meltano_invoker

    @property
    def dagster_name(self) -> str:
        return generate_dagster_name(self.name)

    @property
    def dagster_job(self) -> JobDefinition:
        # We need to import the `meltano_resource` here to prevent circular imports.
        from dagster_meltano.meltano_resource import meltano_resource

        @job(
            name=self.dagster_name,
            description=f"Runs the `{self.name}` job from Meltano.",
            resource_defs={"meltano": meltano_resource},
        )
        def dagster_job():
            meltano_task_done = None
            for task in self.tasks:
                meltano_task_op = task_op_factory(task)
                if meltano_task_done:
                    meltano_task_done = meltano_task_op(meltano_task_done)
                else:
                    meltano_task_done = meltano_task_op()

        return dagster_job



"""Meltano extension SDK base class and supporting methods."""
from __future__ import annotations

import sys
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Any

import structlog
import yaml
from devtools.prettier import pformat
from meltano.edk import models


class DescribeFormat(str, Enum):
    """The currently supported Describe output formats."""

    text = "text"
    json = "json"
    yaml = "yaml"


class ExtensionBase(metaclass=ABCMeta):
    """Basic extension interface that must be implemented by all extensions."""

    def pre_invoke(self, invoke_name: str | None, *invoke_args: Any) -> None:
        """Called before the extension is invoked.

        Args:
            invoke_name: The name of the command that will be passed to invoke.
            *invoke_args: The arguments that will be passed to invoke.
        """
        pass

    def initialize(self, force: bool = False) -> None:
        """Initialize the extension.

        This method is called on-demand by the user to initialize the extension.
        Extensions are not required to implement this method, and may no-op.

        Args:
            force: If True, force initialization.
        """
        pass

    @abstractmethod
    def invoke(self, command_name: str | None, *command_args: Any) -> None:
        """Invoke method.

        This method is called when the extension is invoked.

        Args:
            command_name: The name of the command to invoke.
            *command_args: The arguments to pass to the command.
        """
        pass

    def post_invoke(self, invoked_name: str | None, *invoked_args: Any) -> None:
        """Called after the extension is invoked.

        Args:
            invoked_name: The name of the command that was invoked.
            *invoked_args: The arguments passed to the command that was invoked.
        """
        pass

    @abstractmethod
    def describe(self) -> models.Describe:
        """Describe method.

        This method should describe what commands & capabilities the extension provides.

        Returns:
            Description: A description of the extension.
        """
        pass

    def describe_formatted(
        self, output_format: DescribeFormat = DescribeFormat.text
    ) -> str:
        """Return a formatted description of the extensions commands and capabilities.

        Args:
            output_format: The output format to use.

        Returns:
            str: The formatted description.
        """
        if output_format == DescribeFormat.text:
            return pformat(self.describe())
        elif output_format == DescribeFormat.json:
            return self.describe().json(indent=2)
        elif output_format == DescribeFormat.yaml:
            # just calling describe().dict() and dumping that to yaml yields a yaml that
            # is subtly different to the json variant in that it you have an additional
            # level of nesting.
            return yaml.dump(
                yaml.safe_load(self.describe().json()), sort_keys=False, indent=2
            )

    def pass_through_invoker(
        self, logger: structlog.BoundLogger, *command_args: Any
    ) -> None:
        """Pass-through invoker.

        This method is used to invoke the wrapped CLI with arbitrary arguments.
        Note this method will hard exit the process if an unhandled exception is
        encountered.

        Args:
            logger: The logger to use in the event an exception needs to be logged.
            *command_args: The arguments to pass to the command.
        """
        logger.debug(
            "pass through invoker called",
            command_args=command_args,
        )
        try:
            self.pre_invoke(None, command_args)
        except Exception:
            logger.exception(
                "pre_invoke failed with uncaught exception, please report to maintainer"
            )
            sys.exit(1)

        try:
            self.invoke(None, command_args)
        except Exception:
            logger.exception(
                "invoke failed with uncaught exception, please report to maintainer"
            )
            sys.exit(1)

        try:
            self.post_invoke(None, command_args)
        except Exception:
            logger.exception(
                "post_invoke failed with uncaught exception, please report to maintainer"  # noqa: E501
            )
            sys.exit(1)



import json
import subprocess
from typing import List


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def generate_dagster_name(input_string) -> str:
    """
    Generate a dagster safe name (^[A-Za-z0-9_]+$.)
    """
    return input_string.replace("-", "_").replace(" ", "_").replace(":", "_")


def generate_dbt_group_name(node_info: dict) -> str:
    """Generate the name of the Dagster asset group a DBT nodes lives in.
    This will be namespaced with an extra group if the DBT models are nested in
    sub-folders.

    Args:
        node_info (dict): The information generated for this node by DBT.

    Returns:
        str: The name of the asset group.
    """
    if len(node_info.get("fqn", [])) >= 3:
        return "_".join(node_info["fqn"][1:-1])

    return "dbt"



import os
from pathlib import Path

from dagster import repository, with_resources
{% if cookiecutter.dbt_plugin -%}
from dagster_dbt import dbt_cli_resource
{% endif -%}
from dagster_meltano import (
    load_assets_from_meltano_project,
    load_jobs_from_meltano_project,
    meltano_resource,
)

MELTANO_PROJECT_DIR = os.getenv("MELTANO_PROJECT_ROOT", os.getcwd())
MELTANO_BIN = os.getenv("MELTANO_BIN", "meltano")
{% if cookiecutter.dbt_plugin %}
DBT_PROJECT_DIR = Path(MELTANO_PROJECT_DIR) / "transform"
DBT_PROFILES_DIR = Path(MELTANO_PROJECT_DIR) / "transform" / "profiles" / "{{ cookiecutter.dbt_plugin }}"
{% endif %}
meltano_jobs = load_jobs_from_meltano_project(MELTANO_PROJECT_DIR)


@repository
def repository():
    return [
        meltano_jobs, 
        with_resources(
            load_assets_from_meltano_project(
                meltano_project_dir=MELTANO_PROJECT_DIR,
                {% if cookiecutter.dbt_plugin -%}
                dbt_project_dir=str(DBT_PROJECT_DIR),
                dbt_profiles_dir=str(DBT_PROFILES_DIR),
                # dbt_target_dir=str(DBT_TARGET_PATH),
                dbt_use_build_command=True,
                {%- endif %}
            ),
            {
                "meltano": meltano_resource,
                {% if cookiecutter.dbt_plugin -%}
                "dbt": dbt_cli_resource.configured(
                    {
                        "project_dir": str(DBT_PROJECT_DIR),
                        "profiles_dir": str(DBT_PROFILES_DIR),
                        # "target_path": DBT_TARGET_PATH,
                        # "dbt_executable": "/dagster-testing/.meltano/transformers/dbt-postgres/venv/bin/dbt",
                    },
                ),
                {%- endif %}
            },
        )
    ]



import asyncio
import json
import logging
import os
import subprocess
from functools import lru_cache
from typing import Dict, List, Optional

from dagster import In, Nothing, get_dagster_logger, job, op, resource

from dagster_meltano.job import Job
from dagster_meltano.meltano_invoker import MeltanoInvoker
from dagster_meltano.schedule import Schedule
from dagster_meltano.utils import Singleton, generate_dagster_name

logger = get_dagster_logger()
STDOUT = 1


class MeltanoResource(metaclass=Singleton):
    def __init__(self, project_dir: str = None, meltano_bin: Optional[str] = "meltano"):
        self.project_dir = project_dir
        self.meltano_bin = meltano_bin
        self.meltano_invoker = MeltanoInvoker(
            meltano_bin,
            log_level="info",  # TODO: Get this from the resource config
        )

    async def load_json_from_cli(self, command: List[str]) -> dict:
        _, logs = await self.meltano_invoker.exec(
            None,
            self.meltano_invoker.log_processor_json,
            command,
        )
        return logs[STDOUT]

    async def gather_meltano_yaml_information(self):
        jobs, schedules = await asyncio.gather(
            self.load_json_from_cli(["job", "list", "--format=json"]),
            self.load_json_from_cli(["schedule", "list", "--format=json"]),
        )
        return jobs, schedules

    @property
    @lru_cache
    def meltano_yaml(self):
        jobs, schedules = asyncio.run(self.gather_meltano_yaml_information())
        return {"jobs": jobs["jobs"], "schedules": schedules["schedules"]}

    @property
    @lru_cache
    def meltano_jobs(self) -> List[Job]:
        meltano_job_list = self.meltano_yaml["jobs"]
        return [Job(meltano_job, self.meltano_invoker) for meltano_job in meltano_job_list]

    @property
    @lru_cache
    def meltano_schedules(self) -> List[Schedule]:
        meltano_schedule_list = self.meltano_yaml["schedules"]["job"]
        schedule_list = [Schedule(meltano_schedule) for meltano_schedule in meltano_schedule_list]
        return schedule_list

    @property
    def meltano_job_schedules(self) -> Dict[str, Schedule]:
        return {schedule.job_name: schedule for schedule in self.meltano_schedules}

    @property
    def jobs(self) -> List[dict]:
        for meltano_job in self.meltano_jobs:
            yield meltano_job.dagster_job

            if meltano_job.name in self.meltano_job_schedules:
                yield self.meltano_job_schedules[meltano_job.name].dagster_schedule


@resource(description="A resource that corresponds to a Meltano project.")
def meltano_resource(init_context):
    # project_dir = init_context.resource_config["project_dir"]
    # return MeltanoResource(project_dir)
    return MeltanoResource()


if __name__ == "__main__":
    meltano_resource = MeltanoResource("/workspace/meltano")
    print(meltano_resource.meltano_yaml)



import json
import logging
import subprocess
from typing import List, Optional, Union

from dagster import AssetsDefinition, JobDefinition, ScheduleDefinition
from dagster_dbt import load_assets_from_dbt_project

from dagster_meltano.meltano_resource import MeltanoResource
from dagster_meltano.utils import generate_dbt_group_name


def load_jobs_from_meltano_project(
    meltano_project_dir: Optional[str],
) -> List[Union[JobDefinition, ScheduleDefinition]]:
    """This function generates dagster jobs for all jobs defined in the Meltano project. If there are schedules connected
    to the jobs, it also returns those.

    Args:
        project_dir (Optional[str], optional): The location of the Meltano project. Defaults to os.getenv("MELTANO_PROJECT_ROOT").

    Returns:
        List[Union[JobDefinition, ScheduleDefinition]]: Returns a list of either Dagster JobDefinitions or ScheduleDefinitions
    """
    meltano_resource = MeltanoResource(meltano_project_dir)
    meltano_jobs = meltano_resource.jobs

    return list(meltano_jobs)


def load_assets_from_meltano_project(
    meltano_project_dir: str,
    dbt_project_dir: Optional[str] = None,
    dbt_profiles_dir: Optional[str] = None,
    dbt_target_dir: Optional[str] = None,
    dbt_use_build_command: bool = True,
) -> List[AssetsDefinition]:
    """This function generates all Assets it can find in the supplied Meltano project.
    This currently includes the taps and dbt assets.

    Args:
        project_dir (Optional[str], optional): The location of the Meltano project. Defaults to os.getenv("MELTANO_PROJECT_ROOT").

    Returns:
        List[AssetsDefinition]: Returns a list of all Meltano assets
    """
    # meltano_resource = MeltanoResource(meltano_project_dir)
    meltano_assets = []
    # meltano_assets = [extractor.asset for extractor in meltano_resource.extractors]

    if dbt_project_dir:
        dbt_assets = load_assets_from_dbt_project(
            project_dir=dbt_project_dir,
            profiles_dir=dbt_profiles_dir,
            target_dir=dbt_target_dir,
            use_build_command=dbt_use_build_command,
            node_info_to_group_fn=generate_dbt_group_name,
        )
        meltano_assets += dbt_assets

    return meltano_assets


if __name__ == "__main__":
    load_jobs_from_meltano_project("/workspace/meltano")



"""Various models used to describe extensions."""
from typing import List

from pydantic import BaseModel


class Command(BaseModel):
    """Describe a generic runnable command."""

    name: str
    description: str
    commands: List[str] = []
    pass_through_cli: bool = False


class ExtensionCommand(Command):
    """Describes an extension command."""

    description = "The extension cli"
    pass_through_cli: bool = False
    commands: List[str] = [
        "describe",
        "invoke",
        "pre_invoke",
        "post_invoke",
        "initialize",
    ]


class InvokerCommand(Command):
    """Describes an invoker style command."""

    description = "The pass through invoker cli"
    pass_through_cli: bool = True
    commands: List[str] = [":splat"]


class Describe(BaseModel):
    """Describes what commands and capabilities the extension provides."""

    commands: List[Command]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dagster_ext',
 'dagster_meltano',
 'files_dagster_ext',
 'files_dagster_ext.dagster.{{ cookiecutter.project_name }}',
 'meltano',
 'meltano.edk']

package_data = \
{'': ['*'], 'files_dagster_ext': ['dagster/*', 'github/*']}

install_requires = \
['PyYAML>=6.0.0,<7.0.0',
 'click>=8.1.3,<9.0.0',
 'cookiecutter>=2.1.1,<3.0.0',
 'dagit>=1.0,<1.1',
 'dagster-dbt>=0.16,<0.17',
 'dagster>=1.0,<1.1',
 'devtools>=0.9.0,<0.10.0',
 'pydantic>=1.9.0,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'structlog>=21.2.0,<22.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['dagit_invoker = '
                     'dagster_ext.pass_through:pass_through_cli_dagit',
                     'dagster_extension = dagster_ext.main:app',
                     'dagster_invoker = '
                     'dagster_ext.pass_through:pass_through_cli_dagster']}

setup_kwargs = {
    'name': 'dagster-ext',
    'version': '0.0.1a5',
    'description': '`dagster-ext` is a Meltano utility extension.',
    'long_description': 'None',
    'author': 'Jules Huisman',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import asyncio
import json
import os
import subprocess
from pathlib import Path
from typing import IO, Any, Callable, List, Optional, Tuple, Union

import structlog
from dagster import get_dagster_logger

# log = structlog.get_logger()
log = get_dagster_logger()


class MeltanoInvoker:
    """Invoker utility class for invoking subprocesses."""

    def __init__(
        self,
        bin: str = "meltano",
        cwd: str = None,
        log_level: str = "info",
        env: Optional[dict[str, any]] = {},
    ) -> None:
        """Minimal invoker for running subprocesses.

        Args:
            bin: The path/name of the binary to run.
            cwd: The working directory to run from.
            env: Env to use when calling Popen, defaults to current os.environ if None.
        """
        self.bin = bin
        self.cwd = cwd
        self.env = {
            **os.environ.copy(),
            "MELTANO_CLI_LOG_CONFIG": Path(__file__).parent / "logging.yaml",
            "MELTANO_CLI_LOG_LEVEL": log_level,
            "DBT_USE_COLORS": "false",
            "NO_COLOR": "1",
            **env,
        }

    def run(
        self,
        *args: Union[str, bytes, os.PathLike[str], os.PathLike[bytes]],
        stdout: Union[None, int, IO] = subprocess.PIPE,
        stderr: Union[None, int, IO] = subprocess.STDOUT,
        text: bool = True,
        **kwargs: Any,
    ) -> subprocess.Popen:
        """Run a subprocess. Simple wrapper around subprocess.run.

        Note that output from stdout and stderr is NOT logged automatically. Especially
        useful when you want to run a command, but don't care about its output and only
        care about its return code.

        stdout and stderr by default are set up to use subprocess.PIPE. If you do not
        want to capture io from the subprocess use subprocess.DEVNULL to discard it.

        The Invoker's at env and cwd are used when calling `subprocess.run`. If you want
        to override these you're likely better served using `subprocess.run` directly.

        Lastly note that this method is blocking AND `subprocess.run` is called with
        `check=True`. This means that if the subprocess fails a `CalledProcessError`
        will be raised.

        Args:
            *args: The arguments to pass to the subprocess.
            stdout: The stdout stream to use.
            stderr: The stderr stream to use.
            text: If true, decode stdin, stdout and stderr using the system default.
            **kwargs: Additional keyword arguments to pass to subprocess.run.

        Returns:
            The completed process.
        """
        return subprocess.Popen(
            [self.bin, *args],
            cwd=self.cwd,
            env=self.env,
            stdout=stdout,
            stderr=stderr,
            text=text,
            **kwargs,
        )

    @staticmethod
    async def _log_stdio(reader: asyncio.streams.StreamReader, log_type: str) -> None:
        """Log the output of a stream.
        Args:
            reader: The stream reader to read from.
        """
        while True:
            if reader.at_eof():
                break
            data = await reader.readline()
            log.info(data.decode("utf-8").rstrip())
            await asyncio.sleep(0)

    @staticmethod
    async def log_processor_json(reader: asyncio.streams.StreamReader, log_type: str) -> dict:
        if log_type != "stdout":
            return

        lines = b""

        while True:
            if reader.at_eof():
                break

            line = await reader.readline()
            lines += line

        return json.loads(lines)

    async def exec(
        self,
        sub_command: Union[str, None] = None,
        log_processor: Optional[Callable] = None,
        *args: Union[str, bytes, os.PathLike[str], os.PathLike[bytes]],
    ) -> Tuple[asyncio.subprocess.Process, List[any]]:
        popen_args = []
        if sub_command:
            popen_args.append(sub_command)
        if args:
            popen_args.extend(*args)

        process = await asyncio.create_subprocess_exec(
            self.bin,
            *popen_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.cwd,
            env=self.env,
        )

        log_processor_function = log_processor or self._log_stdio

        logs = await asyncio.gather(
            asyncio.create_task(log_processor_function(process.stderr, log_type="stderr")),
            asyncio.create_task(log_processor_function(process.stdout, log_type="stdout")),
            return_exceptions=True,
        )

        for log in logs:  # raise first exception if any
            if isinstance(log, Exception):
                raise log

        await process.wait()
        return process, logs

    def run_and_log(
        self,
        sub_command: Union[str, None] = None,
        log_processor: Optional[Callable] = None,
        *args: Union[str, bytes, os.PathLike[str], os.PathLike[bytes]],
    ) -> None:
        """Run a subprocess and stream the output to the logger.

        Note that output from stdout and stderr IS logged. Best used when you want
        to run a command and stream the output to a user.

        Args:
            sub_command: The subcommand to run.
            log_processor: Gets called for each log line that is being processed.
            *args: The arguments to pass to the subprocess.

        Raises:
            CalledProcessError: If the subprocess failed.
        """
        process, logs = asyncio.run(self.exec(sub_command, log_processor, *args))
        if process.returncode:
            raise subprocess.CalledProcessError(process.returncode, cmd=self.bin, stderr=None)



"""Dagster cli entrypoint."""

import os
import sys
from typing import List

import structlog
import typer

from dagster_ext.extension import Dagster
from meltano.edk.extension import DescribeFormat
from meltano.edk.logging import default_logging_config, parse_log_level

APP_NAME = "Dagster"

log = structlog.get_logger(APP_NAME)

ext = Dagster()

typer.core.rich = None  # remove to enable stylized help output when `rich` is installed
app = typer.Typer(
    name=APP_NAME,
    pretty_exceptions_enable=False,
)


@app.command()
def initialize(
    ctx: typer.Context,
    force: bool = typer.Option(False, help="Force initialization (if supported)"),
) -> None:
    """Initialize the Dagster plugin."""
    try:
        ext.initialize(force)
    except Exception:
        log.exception(
            "initialize failed with uncaught exception, please report to maintainer"
        )
        sys.exit(1)


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def invoke(ctx: typer.Context, command_args: List[str]) -> None:
    """Invoke the plugin.

    Note: that if a command argument is a list, such as command_args,
    then unknown options are also included in the list and NOT stored in the
    context as usual.
    """
    command_name, command_args = command_args[0], command_args[1:]
    log.debug(
        "called", command_name=command_name, command_args=command_args, env=os.environ
    )
    ext.pass_through_invoker(log, command_name, *command_args)


@app.command()
def describe(
    output_format: DescribeFormat = typer.Option(
        DescribeFormat.text, "--format", help="Output format"
    )
) -> None:
    """Describe the available commands of this extension."""
    try:
        typer.echo(ext.describe_formatted(output_format))
    except Exception:
        log.exception(
            "describe failed with uncaught exception, please report to maintainer"
        )
        sys.exit(1)


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    log_level: str = typer.Option("INFO", envvar="LOG_LEVEL"),
    log_timestamps: bool = typer.Option(
        False, envvar="LOG_TIMESTAMPS", help="Show timestamp in logs"
    ),
    log_levels: bool = typer.Option(
        False, "--log-levels", envvar="LOG_LEVELS", help="Show log levels"
    ),
    meltano_log_json: bool = typer.Option(
        False, "--meltano-log-json",
        envvar="MELTANO_LOG_JSON",
        help="Log in the meltano JSON log format"
    ),
) -> None:
    """Simple Meltano extension that wraps the dagster CLI."""
    default_logging_config(
        level=parse_log_level(log_level),
        timestamps=log_timestamps,
        levels=log_levels,
        json_format=meltano_log_json
    )



from dagster import resource

from dagster_meltano.generation import (
    load_assets_from_meltano_project,
    load_jobs_from_meltano_project,
)
from dagster_meltano.meltano_resource import MeltanoResource, meltano_resource



"""Recommended logging configuration for extensions."""
from __future__ import annotations

import logging
import os
import sys

import structlog

LEVELS = {  # noqa: WPS407
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

DEFAULT_LEVEL = "info"


def parse_log_level(log_level: dict[str, int]) -> int:
    """Parse a level descriptor into an logging level.

    Args:
        log_level: level descriptor.

    Returns:
        int: actual logging level.
    """
    return LEVELS.get(log_level.lower(), LEVELS[DEFAULT_LEVEL])


def default_logging_config(
    level: int = logging.INFO,
    timestamps: bool = False,
    levels: bool = False,
    json_format: bool = False,
) -> None:
    """default/demo structlog configuration.

    Args:
        level: logging level.
        timestamps: include timestamps in the log.
        levels: include levels in the log.
        json_format: if True, use JSON format, otherwise use human-readable format.
    """
    processors = []
    if timestamps:
        processors.append(structlog.processors.TimeStamper(fmt="iso"))
    if levels:
        processors.append(structlog.processors.add_log_level)

    processors.extend(
        [
            # If log level is too low, abort pipeline and throw away log entry.
            structlog.stdlib.filter_by_level,
            # If the "stack_info" key in the event dict is true, remove it and
            # render the current stack trace in the "stack" key.
            structlog.processors.StackInfoRenderer(),
            # If the "exc_info" key in the event dict is either true or a
            # sys.exc_info() tuple, remove "exc_info" and render the exception
            # with traceback into the "exception" key.
            structlog.processors.format_exc_info,
            # If some value is in bytes, decode it to a unicode str.
            structlog.processors.UnicodeDecoder(),
            structlog.processors.ExceptionPrettyPrinter(),
            # Render the final event dict as JSON.
            structlog.processors.JSONRenderer()
            if json_format
            else structlog.dev.ConsoleRenderer(colors=False),
        ]
    )

    structlog.configure(
        processors=processors,
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )

    logging.basicConfig(
        format="%(message)s",
        stream=sys.stderr,
        level=level,
    )


def pass_through_logging_config() -> None:
    """Pass-through logging configuration.

    Setups a logging config using the LOG_LEVEL, LOG_TIMESTAMPS, LOG_LEVELS,
    and MELTANO_LOG_JSON env vars.
    """
    log_level = os.environ.get("LOG_LEVEL", "INFO")
    log_timestamps = os.environ.get("LOG_TIMESTAMPS", False)
    log_levels = os.environ.get("LOG_LEVELS", False)
    meltano_log_json = os.environ.get("MELTANO_LOG_JSON", False)

    default_logging_config(
        level=parse_log_level(log_level),
        timestamps=log_timestamps,
        levels=log_levels,
        json_format=meltano_log_json,
    )
